/*
    dp : An implementation of the davis puttnam procedure for solving
         propositional satisfiability problems.
    Copyright (C) 1993  Peter Barth

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* ------------------------------------------------------------------
   DP.C
* -------------------------------------------------------------------
* Auth: pb
* Purp: A Davis Putnam procedure implementation
* Hist:==============================================================
* Date         Name    What has been done ?!
* -------------------------------------------------------------------
* 20.8.92      pb      creation                
* 20.1.93      pb      faster,cleaned up        
* 10.4.93      pb      single linked lists for main loop
* 12.5.93      pb      slightly better performance and cleaner code
                       by detecting new unit_variables immediately on
                       creation.
		       ( No need to check for empty clause, implicit
		         when two compl. unit_literals are found)
		       Unfortunately this implies an ugly check whether the 
                       initial problem description has a unit_clause.
* 19.5.93      pb      split the sources
* 21.5.93      pb      new heuristics h3, h4
* 23.5.93      pb      new option -s to show solution
* 25.5.93      pb      new heuristic h5; new option -0
* ---------------------------------------------------------------- */

#include "stdlib.h"
#include <ctype.h>
#include <math.h>
#include "datadefs.h"
#include "heuristics.h"
#include "dpio.h"


/**************************************************************************/
/* DO THE JOB                                                             */
/**************************************************************************/

#define FAIL -1
#define OK 1
#define SATISFIABLE 2

int get_ucv(silist av,int* clause)
{
  int absclause = abs(*clause);

  do {  /* assume clause_variables and active variables to be sorted */
    if (av->key < absclause)
      av = av->next;
    else if (av->key > absclause) 
      absclause = abs(*++clause);
    else /* if (av->key == absclause) */
      return(*clause);
#ifdef CHECK
  } while ((av->next != av) && *clause != 0); 
  error("Sorry, a BUG !!! Send me problem !");
#else
  } while (1);
#endif
  exit(-1);
}

int update(int variable,silist av,silist ac)
{
  silist av_act,ac_act;
  int* l;
  int ref,absref;
  int paclause,clauseid;
  int absvariable;

DO_VERBOSE(17,printf("\nupda_start: ");write_a_v(av);printf("  ");write_a_c(ac);)

  /* init ucvs stack */
  ucvsptr = ucvs+1;
  *ucvs = variable;
  *(ucvs+1) = 0;
  while (ucvsptr != ucvs)  /* as long as unit variables */
    {
      av_act = av;
      if (verbosity >= 3) printf("u");
      no_ures++;
      variable = *(--ucvsptr);
      absvariable = abs(variable);
      /* delete variable from av */
      while(av_act->next->key != absvariable)
	av_act = av_act->next;
      si_fastdelnext(av_act);

DO_VERBOSE(14,printf("\n  pruning %i  ",variable<0 ? variable+1:variable -1);)
DO_VERBOSE(20,printf("\nprun_start: ");write_a_v(av);printf("  ");write_a_c(ac);)
      l = lit_to_clauses[absvariable];
      /* l is the list of clauses where variable occurs in */
      ac_act = ac;
      do {
	clauseid = GETtaclauseid(paclause = ac_act->next->key);
	absref = abs(ref = *l);
    
	DO_VERBOSE(21,\
	  printf(" \n   : checking %i with %i    ",\
          abs(*l),GETtaclauseid(paclause));)
	if (absref > clauseid) 
	  ac_act = ac_act->next;
	else if (absref < clauseid) 
	  l++;
	else /* if (absref == clauseid) */
	  {
	    /* efficient for: 
	       if ((ref > 0 && variable > 0) || (ref < 0 && variable < 0)) */
	    if (!((ref ^ variable) & 0x8000)) 
	      {
		si_fastdelnext(ac_act); /* the clause is satisfied */
		if (ac->next->next == ac->next) 
		  return SATISFIABLE;
		DO_VERBOSE(15,printf("\n        del_clause %i ",abs(ref));)
	      }
	    else
	      {
		DO_VERBOSE(15,printf("\n        del_literal in clause % i",abs(ref));)
		switch (GETtaclauselen(paclause))
		  {
#ifdef CHECK
		  case 1 : 
		    error("Sorry, a BUG !!! Send me problem !");
#endif
		  case 2 :                /* new unit clause derived */
		    { 
		      int newucv = get_ucv(av->next,clause_to_lits[absref]);
		      ucvscmpptr = ucvs;
		      DO_VERBOSE(15,printf(" found %i",\
                         newucv<0 ? newucv+1 : newucv-1);)
		      while (ucvscmpptr != ucvsptr)
			{
			  if (*ucvscmpptr == newucv)  /* got it already */
			    { newucv = 0;break; }
			  else if (*ucvscmpptr == -newucv) /* v and -v -> fail */
			    return FAIL;
			  else
			    ucvscmpptr++;
			}
		      /* put new unit_variable on the stack, del clause */
		      if (newucv)
			{ *ucvsptr++ = newucv;*ucvsptr = 0;}
		    }
		    si_fastdelnext(ac_act); /* its a new unit_clause */
		    if (ac->next->next == ac->next) 
		      return SATISFIABLE;
		    break;
		  default :          /* len of clause -1 */
		    ac_act->next->key--; /* len is stored on least sig. bits */
		    ac_act = ac_act->next;
		  }
	      }
	    l++; /* This is updated */
	  }
      } while (*l && (ac_act->next->next != ac_act->next)); 
    }
DO_VERBOSE(17,printf("\nupda_done : ");write_a_v(av);printf("  ");write_a_c(ac);)
  return OK;
}

int solve(silist,silist); /* we need this prototype */

int solve_var(int variable,silist av,silist ac)
{
  switch(update(variable,av,ac)) 
    {
    case FAIL : return FAIL;break;
    case SATISFIABLE : return SATISFIABLE;break;
    default: /* OK */
      return(solve(av,ac));
      break;
    }
}

int solve(silist av,silist ac)
{
  silist cp_av = NULL;
  silist cp_ac = NULL;
  int choice_variable;

  if (ac->next->next == ac->next)
    return SATISFIABLE;
DO_VERBOSE(16,printf("\nsolv:");write_a_v(av);printf("  ");write_a_c(ac);)
  /* copy the active lists */
  cp_av = si_fastcopy(av);
  cp_ac = si_fastcopy(ac);
 
  choice_variable = choose_variable(cp_av,cp_ac);
DO_VERBOSE(19,printf("\ncopy_active: ");write_a_v(cp_av);write_a_c(cp_ac);\
printf("\n org_active: ");write_a_v(av);write_a_c(ac);)
DO_VERBOSE(10,printf("\nchoice on %i :",choice_variable < 0 ? choice_variable+1 : choice_variable-1);)
  if (verbosity >= 2)
    { printf("."); fflush(stdout); }
  no_nodes++; /* statistics */
  if (solve_var(choice_variable,cp_av,cp_ac) == SATISFIABLE)
    {
      free(cp_av);free(cp_ac);
      solution = l_dinsb(solution,(void *) choice_variable); /* collect choices */
      return SATISFIABLE;
    }
  if (verbosity >= 2)
    { printf("B"); fflush(stdout); }
  free(cp_av);free(cp_ac);
  no_nodes++; /* statistics */
DO_VERBOSE(10,printf("\nbacktrackchoice on %i :",-choice_variable < 0 ? -choice_variable+1 : -choice_variable-1);)
  if (solve_var(-choice_variable,av,ac) == SATISFIABLE)
    {
      solution = l_dinsb(solution,(void *) -choice_variable); /* collect choices */
      return SATISFIABLE;
    }
  else
    return FAIL;
}

/* The whole stuff find_ucv unit_resolution and init_upd 
 * are only needed for initial simplification in order to
 * prune away inital unit_clauses and for constructing
 * the solution if the problem is found satisfiable
 */
int find_ucv(silist av,silist ac)
{
  int* clause;
  av = av->next;
  ac = ac->next;

  for(;;) 
    if (ac->next == ac)                      /* no unit clause */
      return 0;
    else if (GETtaclauselen(ac->key)  == 1)  /* got one */
      break;
    else
      ac = ac->next;
DO_VERBOSE(11,printf("\n  : unit_clause is %i",GETtaclauseid(ac->key));)
  /* clause is the clause we search for */
  clause = clause_to_lits[GETtaclauseid(ac->key)]; 
  do {  /* assume clause_variables and active variables to be sorted */
    int avkey = av->key;  
    int vcl = *clause; /* assume clause is valid */

    if (avkey == abs(vcl))
      return(vcl);
    else if (avkey < abs(vcl))
      av = av->next;
    else /* avkey > abs(vcl) */
      clause++;
  } while ((av->next != av) && *clause != 0); 
#ifdef CHECK
  write_a_v(av);printf("   ");write_a_c(ac);
  error("Sorry, a BUG !!! Send me problem !");
#endif
  exit(-1);
}

int init_upd(int variable,silist av,silist ac)
{
  silist av_act = av;
  silist ac_act = ac;
  int* l;
  int i;
  int ref,absref;
  int paclause;
  int absvariable = abs(variable);

DO_VERBOSE(12,printf("\nupda_start: ");write_a_v(av);printf("  ");write_a_c(ac);)
  /* delete variable from active variable list (assume av = l_first(av)) */
  while(av_act->next->key != absvariable)
     av_act = av_act->next;
  si_fastdelnext(av_act);
  av_act = av;
  /* refresh clause list */
  l = lit_to_clauses[absvariable];
  i = 0;
  /* Now l is the list of clauses where variable occurs in */
  do {
    paclause = ac_act->next->key;
    absref = abs(ref = l[i]);
    
    if (absref == GETtaclauseid(paclause)) /* Treffer */
    {
      /* efficient for: if ((ref > 0 && variable > 0) || (ref < 0 && variable < 0)) */
      if (!((ref ^ variable) & 0x8000))
      {
DO_VERBOSE(20,\
printf("\n    : checking %i with %i   del_clause",abs(ref),GETtaclauseid(paclause));)
	si_fastdelnext(ac_act);
      }
      else
      {
        if (GETtaclauselen(paclause) == 1)   /* empty clause derived */
	  return FAIL;
DO_VERBOSE(20, \
printf("\n    : checking %i with %i   del_literal",abs(ref),GETtaclauseid(paclause));)
	ac_act->next->key--; /* Note that len is stored on least sig bits */
        if (ac_act->next->next == ac_act->next) 
	  break;
	else 
	  ac_act = ac_act->next;
      }
      i++; /* This is updated */
    }
    else if (absref < GETtaclauseid(paclause))
      i++;
    else /* if (absref < GETtaclauseid(paclause)) */
    {
      if (ac_act->next->next == ac_act->next)
	break;
      else
	ac_act = ac_act->next;
    }
  } while (l[i]); /*  ac_act at end catched with breaks (speed) */
DO_VERBOSE(12,printf("\nupda_done : ");write_a_v(av);printf("  ");write_a_c(ac);)
  return OK;
}

int unit_resolution(silist av,silist ac)
{
  int unit_clause_variable;
  
  do {
DO_VERBOSE(10,printf("\nures:");write_a_v(av);printf("  ");write_a_c(ac);)
    if ((unit_clause_variable = find_ucv(av,ac)) == 0)
      break;
    else
    {
      if (verbosity >= 3) printf("u");
      no_ures++; /* statistics */
      /* collect all variables */
      solution_all = l_dinsb(solution_all,(void *) unit_clause_variable); 
      if (init_upd(unit_clause_variable,av,ac) == FAIL)
        return FAIL;
    }
  } while(1); /* (ac->next != ac->next->next); (implicit, find_ucv will fail) */
DO_VERBOSE(10,printf("\nures_done");)
    if (ac->next == ac->next->next)
      return SATISFIABLE;
    else
      return OK;
}

int dp_solve(silist av,silist ac)
{
  /* throw away unit clauses from the initial problem description */
  switch (unit_resolution(av,ac))
    {
    case SATISFIABLE : return SATISFIABLE;
    case FAIL : return FAIL;
      default : return solve(av,ac);
    }
}



/*************************************************************************
 * show the solution                                                      
 * We have to reconstruct it because we don't want to loose speed while
 * solving.
 **************************************************************************/

void show_solution()
{
  list good_choices = l_first(solution);
  silist av = list2silist(active_variables);
  silist ac = list2silist(active_clauses);
  int _verbosity = verbosity;

  verbosity = 0;
  l_delall(solution_all);
  solution_all = NULL;
  unit_resolution(av,ac);
  printf("\nsolution   : ");
  printf("\n    chosen : ");
  while(good_choices)
    {
      int var = (int) l_rdata(good_choices);

      printf("%i ",var > 0 ? var-1 : var +1); /* as we add one in the parser */
      init_upd(var,av,ac);
      unit_resolution(av,ac);
      good_choices = l_next(good_choices);
    }
  solution_all = l_first(solution_all);
  printf("\n   chained : ");
  while(solution_all) 
    { 
      int var = (int) l_rdata(solution_all);

      printf("%i ",var > 0 ? var-1 : var +1); /* as we add one in the parser */
      solution_all = l_next(solution_all);
    } 
#ifdef CHECK
  if (ac->next->next != ac->next)
    error("Solution seems not to be one !");
#endif
  free_silist(ac);
  free_silist(av);
  verbosity = _verbosity;
}


/*************************************************************************/
/*                           D P H E L P                                 */
/*************************************************************************/

void dphelp()
{
  printf("This is an implementation of the well known Davis Putnam Procedure \n");
  printf("for solving propositional satisfiability problems\n");
  printf("The following options are available \n");
  printf("\n");
  printf("  -H       prints this message \n\n");
  printf("  -h<n>    chose a variable selection heuristic \n");
  printf("    -h0      no heuristic -> first occurrence; first try (default) \n");
  printf("    -h1      Jeroslow Wang heuristic\n");
  printf("    -h2      var with most often occurrence with most often sign \n");
  printf("    -h3      variant of h1, (combines h1 and h2) \n");
  printf("  -tn      timing 0: no timing. 1: overall cputime. 2: init <-> solve \n");
  printf("  -f<name> read input from file <name> (else from stdin) \n");
  printf("  -v<n>    verbosity level 0: nothing. 1: +init+result 2:+choice+backt. \n");
  printf("                           3: +unitres 4: +problem description\n");
  printf("                           higher for debugging \n");
  printf("  -s       show solution if problem is satsifiable \n");
  printf("  -0       forbids 0 as literal and allows 0 to be end of clause. \n");
  printf("           Additionally EOLN is no longer end of clause \n");
  printf("\n");
  printf(" Name   : dp \n");
  printf(" Version: 0.2 // 21.5.93 \n");
  printf(" Author : Peter Barth \n");
  printf(" Email  : barth@mpi-sb.mpg.de \n");
  printf("\n");
  exit(-1);
}


/*************************************************************************/
/*                              M A I N                                  */
/*************************************************************************/

int main(int argc,char** argv)
{
  FILE *file = NULL;
  int res;
  int tmp_time = 0;
  int start_time = 0;
  int read_time = 0;
  int init_time = 0;
  int solve_time = 0;
  int end_time = 0;

  if (argc >= 2)
  {
    int i = 1;
    char buffer[60];

    while (argc > i) {
      strcpy(buffer,argv[i]);
      if (buffer[0] != '-')
        error("option don't begin with -");
      switch(buffer[1]) {
        case 'f' : file =fopen(buffer+2,"r");
                   if (file == NULL)
                     error("File not found");
                   break;
        case 'v' : verbosity = atoi(buffer+2);
                   break;
        case 't' : timing = atoi(buffer+2);
                   break;
        case 'h' : 
	           if (isdigit(*(buffer+2)))
		     heuristic = atoi(buffer+2);
	           else
		     error("-h?  which heuristic ?");
                   break;
	case 'H' : dphelp();
	           break;
	case 's' : showsolution = 1;
	           break;
	case '0' : allow_zero = 0;
	           break;
	default  : error("unknown option");
      }
      i++;
    }
  }
  if (file == NULL)
    file = stdin;
  if (timing)
    start_time = get_time();
  read_clauses(file);
  if (timing)
    read_time = get_time()-start_time;
  if (verbosity >= 1)
    { 
      printf("\
dp version 0.2, Copyright (C) 1993 Peter Barth \n\
dp comes with ABSOLUTELY NO WARRANTY; for details look in the file COPYING. \n\
This is free software, and you are welcome to redistribute it \n\
under certain conditions; look in the file COPYING for details. \n");
      printf("type    dp -H    for help \n");
      fflush(stdout); 
    }
  report_problem();
  if (verbosity >= 1)
    { 
      printf("\nInput completed --> Initializing datastructures ... "); 
      fflush(stdout); 
    }
  if (timing)
    tmp_time = get_time();
  init_datastructures();
  if (timing)
    init_time = get_time()-tmp_time;
  report_datastructures();
  if (verbosity >= 1)
    { 
      printf("\nInitializing datastructures completed  --> Solving ... \n"); 
      fflush(stdout); 
    }
  if (timing)
    tmp_time = get_time();
  res = dp_solve(av,ac);
  if (timing) 
    {
      end_time = get_time();
      solve_time = end_time-tmp_time;
    }
  if (res == FAIL) 
    {
      if (verbosity >= 1)
	printf("\n ****** INSATISFIABLE ******");
    }
  else
    if (verbosity >= 1)
      printf("\n ******  SATISFIABLE  ******");
  if (timing == 1)
    printf("\nTime used : %i msec.",end_time-start_time);
  if (timing == 2)
    printf("\nTime used (msecs) : read(%i) init(%i) solve(%i) ",
                                  read_time,init_time,solve_time);
  if (verbosity >= 1)
    printf("\nExplored nodes : %i     Number of unit resolutions : %i ",
	   no_nodes,no_ures);
  if (showsolution && (res == SATISFIABLE))
    show_solution();
  if (verbosity >= 1)
    printf("\nBye ...\n");
  clean_datastructures();
  return(res == FAIL ? 1 : 0);
}



